// $Header: E:/TNG/src/rcs/populat.cpp 1.2 1997/02/22 14:38:06 dbm Exp $

#include <stdlib.h> // for qsort()
#include <stdio.h>
#include "populat.h"
#include "random.h"

////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//	bioPopulation
//
 
bioPopulation::bioPopulation (void)
{
	_size = 0;
	_elite = 1;
	_sorted = FALSE;

	_ptypes = 0;         
	
	_mutationRate = 0.01;
	
	_totalFitness = 0.0;
	_maxFitness = 0.0;
	_minFitness = 1.7E+308;
	_avgFitness = 0.0;
}             

bioPopulation::~bioPopulation (void)
{     
	for (int i = 0; i < _size; i++)
	{
		delete _ptypes[i];
	}

	delete _ptypes;
}             

void
bioPopulation::SetSize (int size)
{
	ASSERT (size > 0);
	ASSERT (_ptypes == 0);
	 
	_size = size;
	_ptypes = new bioPType*[_size];
	
	ASSERT (_ptypes);
	
	for (int i = 0; i < _size; i++)
	{
		_ptypes[i] = GenPType (GenGType());  
		_ptypes[i]->SetID(i);
	}
}             

int
bioPopulation::GetSize (void)
{              
	return _size;
}

void
bioPopulation::SetElite (int elite)
{                                   
	ASSERT (elite >= 0);
	
	_elite = elite;
}

int
bioPopulation::GetElite (void)
{
	return _elite;
}

void
bioPopulation::SetMutationRate (double mrate)
{                   
	// Assertion changed to allow for 0 mutation rate.
	// dbm 97.01.30
	ASSERT (mrate >= 0.0);
	ASSERT (mrate < 1.0);
	
	_mutationRate = mrate;
}

double
bioPopulation::GetMutationRate (void)
{
	return _mutationRate;
}

bioPType*
bioPopulation::GetPType (int idx)
{
	ASSERT (idx >= 0);
	ASSERT (idx < _size);
	
	return _ptypes[idx];
} 

static int
compFitness (const void* a, const void* b)
{
	bioPType*	p1 = *(bioPType**) a;
	bioPType*	p2 = *(bioPType**) b;

	ASSERT_VALID (p1);
	ASSERT_VALID (p2);
		                                        
	float		x = (float) p1->GetFitness();
	float		y = (float) p2->GetFitness();
	
	int			comp = CompareFloats(x, y);

	// printf("compFitness(%s, %s) -> %d\n", p1->GetName(), p2->GetName(), comp);

	return comp;
}

void
bioPopulation::SortByFitness (void)
{
	if (_size == 0)
	{
		return;
	}

	qsort (_ptypes, _size, sizeof(_ptypes[0]), compFitness); 

	_totalFitness = 0;
	_maxFitness = 0;
	_minFitness = 1.7E+308;

	for (int i = 0; i < _size; i++)
	{           
		double			fit = _ptypes[i]->GetFitness();
		_totalFitness += fit;
		
		if (fit > _maxFitness)
		{
			_maxFitness = fit;
		} 
		
		if (fit < _minFitness)
		{
			_minFitness = fit;
		}
	}           
	
	_avgFitness = _totalFitness/_size;
	_sorted = TRUE;
	
	Trace ("%f %f %f %f\n", _minFitness, _maxFitness, _totalFitness, _avgFitness);
}

void
bioPopulation::Breed (void)
{	
	if (_size == 0)
	{
		return;
	}

	if (_sorted == FALSE)
	{
		SortByFitness();
	}

	bioPType**			nextGen = new bioPType*[_size];
   	                                                    
	ASSERT (nextGen);
	
	// Transfer the elite gtypes directly into the next generation.

	int		i;

	for (i = 0; i < _elite; i++)
	{
		nextGen[i] = GenPType(_ptypes[i]->GetGType());
		nextGen[i]->SetID(i);
	}
	   
	// Replace all but the elite with children
	for ( ; i < _size; i++)
	{
		bioGType*		mom = _ChooseParent();
		bioGType*		dad = _ChooseParent();
		bioGType*		dna = mom->Crossover (dad);
		
		// Don't bother mutating unless mutation rate is non-zero.
		// dbm 97.01.30
		if (_mutationRate > 0.0)
		{
			dna->Mutate (_mutationRate);
		}

		nextGen[i] = GenPType(dna);
		nextGen[i]->SetID(i);
	}
	                                                    
	// Extract the elite gtypes so they aren't deleted.
	for (i = 0; i < _elite; i++)
	{
		_ptypes[i]->ExtractGType();
	}

	for (i = 0; i < _size; i++)
	{
		delete _ptypes[i];
	}
	
	delete _ptypes;
	_ptypes = nextGen;

	_sorted = FALSE;
}

//
// Socialize -  similar to Breed() but does not destroy the phenotypes,
//				instead only modifies the genetic material (genotypes)
//				so that identity and memory are not destroyed. The genetic
//				material in this case is interpreted as a mental state.
//
//				DJS 4/12/00

void bioPopulation::Socialize()
{
	if (_size == 0)
	{
		return;
	}

	if (_sorted == FALSE)
	{
		SortByFitness();
	}

	bioGType**			nextGen = new bioGType*[_size];
   	                                                    
	ASSERT (nextGen);
	
	// Transfer the elite gtypes directly into the next generation.

	int		i;

	for (i = 0; i < _elite; i++)
	{
		nextGen[i] = _ptypes[i]->GetGType();
	}
	   
	// Get modified gtypes for all but the elite
	for ( ; i < _size; i++)
	{
		bioGType*		mom = _ChooseParent();
		bioGType*		dad = _ChooseParent();
		bioGType*		dna = mom->Crossover (dad);
		
		// Don't bother mutating unless mutation rate is non-zero.
		// dbm 97.01.30
		if (_mutationRate > 0.0)
		{
			dna->Mutate (_mutationRate);
		}

		nextGen[i] = dna;
	}
	                                                    
	// Set the new gtypes
	for (i = 0; i < _size; i++)
	{
		_ptypes[i]->InsertGType(nextGen[i]);
	}
	
	delete nextGen;

	_sorted = FALSE;
}


bioGType*
bioPopulation::_ChooseParent (void)
{
	double		r = drand();
	double		x = r * _totalFitness;
	double 		sum = 0;
	int			i;
	
	//printf("_ChooseParent: r=%f totalFitness=%f\n", r, _totalFitness);

	for (i = 0; i < _size; i++)
	{
		sum += _ptypes[i]->GetFitness();
		if (sum > x)
		{
			//printf ("_ChooseParent: %d of %d sum=%f x=%f\n", i, _size, sum, x);
			return _ptypes[i]->GetGType();
		}
	}                            
	
	//printf ("_ChooseParentx: %d of %d\n", i, _size);
	return _ptypes[_size-1]->GetGType();
} 

double
bioPopulation::GetMinFitness (void)
{            
	return _minFitness;
} 

double
bioPopulation::GetMaxFitness (void)
{            
	return _maxFitness;
}

double
bioPopulation::GetAvgFitness (void)
{            
	return _avgFitness;
}



