//
//  tngSim.h (for tnglab.101 -- not for use with tngshell)
//
//  Header file for tngSim.cpp
//  Last Revised: 29 June 1998
//

#ifndef TNGSIM_H
#define TNGSIM_H

#include <stdio.h>
#include "tng.h"        // TNG header files

// utility for standard deviation calculation
float stddevpop(float*, int);

class tngScores
{
public:
	ctor			tngScores();

	float			minScore;
	float			avgScore;
	float			maxScore;
	float			totScore;
	float			stddevScore;
};


class tngSimulation : public bioSimulation
{
public:
	ctor              tngSimulation (void);   // Constructor.
	dtor              ~tngSimulation (void);  // Virtual destructor.

	virtual void	  OpenFile (const char* file);// read parameters from file
	virtual void	  SaveFile (const char* file);// write parameters to file
	
	virtual const char*	GetDumpFile();
	virtual void		SetDumpFile(const char* file); // set to "" for stdout
	virtual void		DUMP(char* fmt, ...);	// dump formatted string to dump file

	virtual void      Init (void);            // Hardcode TNG parameters.
	virtual void      InitGen (int gen);      // Initialize each generation.
	virtual void      InitTradeBots ();		  // Set up trading lists and initialize bots.
	virtual void      MatchTraders (void);    // Determine trade partners.
	virtual void      Trade (int);            // Implement trades.
	virtual void      AssessFitness (void);   // Calculate fitness scores.
	virtual void      Dump (void);            // Print global fitness info.
	virtual void      DumpInput(void);		  // Print input parameters.
	virtual void      DumpBots (void);        // Print detailed bot info.
	virtual void      EvolveGen (void);       // Evolve current generation.

	virtual float     DilemmaPayoff (int a, int b);  // Assign PD payoffs.
	virtual void      MediateTrade (tngTradeBot* seller,tngTradeBot* buyer);

	virtual int       GMax (void)             { return _gmax; }
	virtual int       IMax (void)             { return _imax; }
	virtual bioList&  TraderList (void)       { return _traders; }
	virtual bioList&  BuyerList (void)        { return _buyers; }
	virtual bioList&  SellerList (void)       { return _sellers; }
	virtual int       BuyerQuota (void)       { return _buyerQuota; }
	virtual int       SellerQuota (void)      { return _sellerQuota; }
	virtual float     InitExpPayoff (void)    { return _initExpPayoff; }
	virtual float     RefusalPayoff (void)    { return _refusalPayoff; }
	virtual float     InactivePayoff (void)		{ return _inactivePayoff; }
	virtual	float	  ExperienceGain(void)	{return _experienceGain; }

	virtual void	  CalcScores();

protected:
	void			  _PrepareBotsForNextGeneration();

protected:
	tngPopulation     _pureBuyPopulation;  // Subpopulation of pure buyers.
	tngPopulation     _pureSellPopulation; // Subpopulation of pure sellers.
	tngPopulation     _buysellPopulation;  // Subpopulation of buyer-sellers.

	int     _gmax;    // Number of generations.
	int     _imax;    // Number of trade cycles per generation.

	int		_seed;
	float	_mrate;
	int		_fsmStates;
	int		_fsmMemory;

	float	_elitePercentPB;
	float	_elitePercentPS;
	float	_elitePercentBS;

	int		_eliteCountPB;
	int		_eliteCountPS;
	int		_eliteCountBS;


	bioList _traders; // List of all tradebots.
	bioList _buyers;  // List of all buyers (pure buyers and buyer-sellers).
	bioList _sellers; // List of all sellers (buyer-sellers and pure sellers).

	int     _buyerQuota;        // The maximum number of offers a buyer
		                        //   can have outstanding to sellers.
	int     _sellerQuota;       // The maximum number of offers a seller
			                    //   can have on his waiting list.
	float   _initExpPayoff;     // The expected payoff each tradebot first
				                //   associates with each other tradebot.
	float   _refusalPayoff;     // The payoff received by a tradebot
					            //   whenever he has an offer refused.
	float   _inactivePayoff;  // The payoff received by a tradebot if he
						        //   is inactive during a trade cycle.
	float   _bothCoop;          // The payoff received by a tradebot when
							    //   he and his partner cooperate.
	float   _bothDefect;        // The payoff received by a tradebot when
								//   he and his partner defect.
	float   _temptation;        // The payoff received by a tradebot who
								//   defects while his partner cooperates.
	float   _sucker;            // The payoff received by a tradebot who
								//   cooperates while his partner defects.
	int     _traderCount;       // The total number of tradebots.
	int     _buyerCount;        // The number of tradebots that are buyers.
	int     _sellerCount;       // The number of tradebots that are sellers.
	int     _pureBuyerCount;    // The number of pure buyer tradebots.
	int     _pureSellerCount;   // The number of pure seller tradebots.
	int     _buysellCount;      // The number of tradebots that are both
		                        //    buyers and sellers.
	int     _trades;            // Trade count for the current trade cycle.

	int     _printGen;          // The number of trades is printed out for
	int     _printIt;           //    the final _printIt iterations of the
		                        //    final _printGen generations.
	float	_experienceGain;	// Factor applied when updating the initial exp
								// payoff before each generation

protected:
	tngScores		_allScores;
	tngScores		_pbScores;
	tngScores		_psScores;
	tngScores		_bsScores;

	char*			_dumpFile;	// Filename to dump results to, if Null results are dumped to stdout

};

//
// DETAILED PRE AND POST CONDITIONS FOR TNGSIMULATION METHODS.
//

//  ctor   tngSimulation (void);
     /*
     **  Pre:     True
     **
     ** Post:     The constructor for bioSimulation (public to
     **           tngSimulation) is invoked.
     */
//  dtor   ~tngSimulation (void);
     /*
     **  Pre:     True
     **
     ** Post:     _traders, _buyers, and _sellers are emptied of
     **           elements.  
     */
//  virtual void  Init (void);
     /*
     ** Pre:      True.
     **
     ** Post: (1) The user-supplied parameter values in tng.ini
     **           are assigned to the corresponding tng parameters.
     **       (2) The parameter values are printed out, together with
     **           the current time and date.
     **       (3) The size is set for each nonempty subpopulation of
     **           distinct trader types (pure buyers, pure sellers,
     **           and buyer-sellers).  This triggers the construction
     **           of pure buyer, pure seller, and buyer-seller arrays of
     **           pointers to new tngTradeBot instances (tradebots) with
     **           randomly initialized bit-string genotypes, with each
     **           tradebot's FSM initialized with a bit string extracted
     **           from the tradebot's genotype, and with other member
     **           data set at default values.
     */
//  virtual void  InitGen (int g);
     /*
     **  Pre: The user-supplied parameter values in tng.ini fall
     **       within admissable ranges.
     **
     ** Post: (1) The current generation g+1 is printed out.
     **       (2) The trader list is filled in order with pure buyers,
     **           pure sellers, and buyer-sellers.
     **       (3) Buyer/seller lists are filled with buyers and sellers.
     **       (4) Through successive calls to tngTradeBot::Init(),
     **         (a) Each buyer tradebot sets its buyer offer quota to the
     **             value specified in tng.ini.
     **         (b) Each seller tradebot sets its seller acceptance quota
     **             to the value specified in tng.ini.
     **         (c) Each tradebot constructs a trader information list
     **             to hold information about other tradebots.
     **         (d) Each seller tradebot constructs a waiting list to
     **             hold offers from buyer tradebots.
     **         (e) Each tradebot sets the initial expected payoff it
     **             assigns to itself equal to -1 (so it won't trade with
     **             itself) and the initial expected payoff it assigns to
     **             each other tradebot is set equal to the value
     **             specified in tng.ini.
     */
//  virtual void  MatchTraders (void);
     /*
     **  Pre:     Each tradebot associates an expected payoff with
     **           each other tradebot.
     **
     ** Post: (1) A matching outcome (list of buyer-seller trade
     **           partners) is determined.
     **       (2) If any tradebot had a trade offer refused by some
     **           tradebot i during this matching process, it added a
     **           refusal payoff to its total payoff sum, it added 1
     **           to its total payoff count, it added a refusal payoff
     **           to its payoff sum with tradebot i, it added 1 to its
     **           payoff count with tradebot i, it added 1 to its
     **           refusal count with tradebot i, and it updated its
     **           expected payoff for tradebot i.
     **       (3) If any tradebot neither made nor accepted a trade
     **           offer during this matching process, it added an
     **           inactive payoff to its total payoff sum, it added
     **           1 to its total inactive count, and it added 1 to
     **           its total payoff count.
     */
 // virtual void  Trade (int cycle);
     /*
     **  Pre:     A matching outcome is determined.
     **
     ** Post: (1) A trade has been carried out by each pair of
     **           trade partners listed in the matching outcome.
     **       (2) Each time a tradebot received a payoff P in a trade
     **           with another tradebot i, it added P to its total
     **           payoff sum, it added 1 to its total payoff count,
     **           it added P to its payoff sum with tradebot i, it
     **           added 1 to its payoff count with tradebot i, it
     **           recorded tradebot i's trade action, it added 1 to
     **           its trade count with tradebot i, and it updated its
     **           expected payoff for tradebot i.
     **       (3) The total number of trades is recorded for each trade
     **           cycle, and the total number of trades is printed out
     **           for the final PrintIt trade cycles in each of the
     **           final PrintGen generations, where PrintIt and
     **           PrintGen are set in tng.ini.
     */
//  virtual void  AssessFitness (void);
     /*
     **  Pre:     The total payoff count of each tradebot is positive.
     **
     ** Post: (1) The fitness score of each tradebot is assessed.
     **       (2) If this is the final generation of tradebots,
     **           information about each tradebot is printed out.
     **       (3) Each subpopulation of distinct trader types (pure
     **           buyers, pure sellers, and buyer-sellers) is sorted
     **           by fitness scores.
     */
//  virtual void  Dump (void);
     /*
     **  Pre:     Each subpopulation of distinct trader types (pure
     **           buyers, pure sellers, and buyer-sellers) is sorted
     **           by fitness scores.
     **
     ** Post:     The minimum, maximum, and average fitness scores
     **           attained by each subpopulation and by the population
     **           as a whole are printed out.
     */
//  virtual void  EvolveGen (void);
     /*
     **  Pre: (1) The values for the mutation rate and the number of
     **           elite set in tng.ini are admissible.
     **       (2) the current generation of tradebots is sorted
     **           by fitness scores.
     **
     ** Post: (1) The elements of_traders, _buyers, _sellers are
     **           removed.
     **       (2) The current subpopulations of pure buyers, pure
     **           sellers, and buyer-sellers are replaced with new
     **           subpopulations of pure buyers, pure sellers, and
     **           buyer-sellers.
     */
//  virtual float DilemmaPayoff (int a, int b);
     /*
     **  Pre: True
     **
     ** Post: If sgn(a,b) = (+,+), _bothCoop is returned as a's payoff.
     **       If sgn(a,b) = (+,0), _sucker is returned as a's payoff.
     **       If sgn(a,b) = (0,+), _temptation is returned as a's payoff.
     **       If sgn(a,b) = (0,0), _bothDefect is returned as a's payoff.
     */
//  virtual void  MediateTrade (tngTradeBot* seller,tngTradeBot* buyer);
     /*
     **  Pre: (1) A matching outcome exists.
     **       (2) The tradebots seller and buyer are matched as trade
     **           partners by this matching outcome.
     **
     ** Post: (1) An action (cooperate or defect) is retrieved for
     **           both the buyer and the seller from their FSMs.
     **       (2) The trade payoff received by the buyer as a result
     **           of these actions is conveyed to the buyer, along with
     **           the seller's action; and similarly for the seller.
     **       (3) Information about this trade is printed out.
     */
#endif // TNGSIM_H

