//
//  tngBot.h (for tnglab.101)
//
//  Header file for tngBot.cpp and auxiliary class
//  Last Revised:  6 June 1998
//

#ifndef TNGBOT_H
#define TNGBOT_H

#include <stdio.h>
#include "tng.h"           // TNG header files

/////////////////////////////////////////////////////////
//
//  class TraderInfo:  Auxiliary class for tngTradeBot
//
/////////////////////////////////////////////////////////

class TradeHistory
{
public:
    ctor        TradeHistory();
    dtor        ~TradeHistory();

	int			Size()	{ return _size;}
    char&       operator[](int idx);

protected:
    int         _actualSize;	// including padding for expansion
    int         _size;			// valid data
    char*       _tradeStatus;	//0 = no trade, 1 = trade
};


class TraderInfo       // Holds this tradebot's info about another bot.
{
public:
  ctor  TraderInfo (void);    // Constructor.

  int			TradingStreak();	// Number of most recent consecutive trade cycles
								// in which this tradebot and bot traded
								// (positive value) or didn't trade (negative value)

public:
  tngTradeBot*  trader;       // Which bot is this info about?

  float         lastPayoff;   // This tradebot's last payoff from an
                              //   interaction with bot.
  float         sumPayoffs;   // This tradebot's payoff sum from
                              //   interactions with bot.
  int           numPayoffs;   // This tradebot's payoff count (no. of
                              //   payoffs) from interactions with bot.
  int           numRefusals;  // The number of offers this tradebot has
                              //   had refused by this bot.
  float         expPayoff;    // The expected payoff this tradebot
                              //   currently associates with bot.
  float         initExpPayoff;    // The expected payoff this tradebot
                              //   associated with bot at the start of generation.
  bioBitVector  actionHistory; // Record of bot's actions in trades
                              //   with this tradebot (bitstring).
  TradeHistory	tradeHistory; // Record of whether this tradebot traded with
							  // the bot in each trade cycle
  
  int           numTrades;    // This tradebot's trade count (number
                              //   of previous trades) with bot.
  int           offerStatus;  // This tradebot's offer status with bot
                              //   (no offer made to bot, offer made
                              //    to bot, offer refused by bot).
  bioBitVector  state;        // This tradebot's current state with
                              //    respect to bot.


  enum          {NO_OFFER, OFFER_MADE, OFFER_REJECTED};
};



//////////////////////////////////////
//
//   class tngTradeBot
//
/////////////////////////////////////

class tngTradeBot : public bioPType
{
public:
  ctor          tngTradeBot (bioGType* dna); // Constructor
  dtor          ~tngTradeBot (void);         // Virtual destructor

  static void   SetMemory (int m);           // Trade strategy methods.
  static void   SetStates (int s);
  static void   SetStateBits (int b);
  static int    FSMSize (void);
  virtual void  InitProgram (void);

  virtual void	PrepareForNextGeneration();

  virtual void  Init (tngSimulation* sim,    // InitGen() methods.
    bioList& traders, int maxOffers, int maxAccepts);

  virtual void  Look (void) {}               // MatchTraders() methods.
  virtual void  PrepareToMakeOffers (void);
  virtual void  PrepareToAcceptOffers (void);
  virtual int   SubmitOffers (void);
  virtual void  TakeOffer (tngTradeBot* bot);
  virtual void  AcceptOffers (void);
  virtual void  OfferRejected (tngTradeBot* bot);

  virtual void  PlayPD (void);               // Trade() methods.
  virtual int   GetAction (tngTradeBot* bot);
  virtual void  Buy (tngTradeBot* bot, float lastPayoff, int action);
  virtual void  Sell (tngTradeBot* bot, float lastPayoff, int action);
  virtual void  InactiveCheck (void) ;
  virtual void  UpdateExp (TraderInfo* info);

  virtual float CalcFitness (void);          // Tradebot output.
  virtual void  Dump (void);

  virtual TraderInfo*  GetTraderInfo (tngTradeBot* bot);
  virtual int  IsBuyer (void)    { return (_maxOffers > 0); }
  virtual int  IsSeller (void)   { return (_maxAccepts > 0); }

protected:
  virtual float	_NewInitExpPayoff(TraderInfo* info);



protected:
  tngSimulation* _sim;

  static int     _Memory;     // Number of bits used in FSM coding
                              //  for past trade partner actions.
  static int     _StateBits;  // Number of bits used in FSM coding
                              //  for the number of internal states.
  int    _maxOffers;          // Quota on offers outstanding.
  int    _maxAccepts;         // Quota on offers accepted.
  int    _traderCount;        // Total number of tradebots.
  //int    _action;             // Action taken in a trade. (DJS commented out 7/31/2000, unreferenced) 
  int    _offers;             // No. of offers currently outstanding.
  int    _offersMade;         // No. of offers made (this trade cycle).
  int    _accepts;            // No. offers accepted (this trade cycle).
  int    _inactiveCount;    // Total inactive count (all trade cycles).
  int    _numTotalPayoffs;    // Total payoff count (all trade cycles).
  float  _sumTotalPayoffs;    // Total payoff sum (all trade cycles).

  TraderInfo*    _traderList; // Holds info about each other tradebot.
  TraderInfo*    _waitList;   // Holds offers from other tradebots.


};

//
// DETAILED PRE AND POST CONDITIONS FOR TNGTRADEBOT METHODS.
//

//  ctor  tngTradeBot (bioGType* dna);   // Constructor
   /*
   **  Pre:     True
   **
   ** Post: (1) dna has been passed to the bioPtype constructor.
   **       (2) The tradebot's member data are initialized with
   **           default values.
   **       (3) As a result of a call to InitProgram():
   **           (a) A new FSM has been constructed for this tradebot
   **               with memory and number of internal states
   **               determined in accordance with the specifications
   **               in tng.ini.
   **           (b) The FSM is initialized with an extracted genotype.
   */
//  dtor  ~tngTradeBot (void);           // Destructor.
   /*
   **  Pre:     True
   **
   ** Post:     Arrays and other constructs newly created for this
   **           tradebot are deleted.  
   */
//  static void   SetMemory (int m);
   /*
   **  Pre:     m is a nonnegative integer.
   **
   ** Post:     The FSM memory is set to m.
   */
//  static void   SetStates (int s);
   /*
   **  Pre:     s is a positive integer
   **
   ** Post:     The number of bits that code for s is set to
   **           the smallest integer not less than log(s)/log(2).
   */
//  static void   SetStateBits (int b);
   /*
   **  Pre:     b is an integer from 1 through 15.
   **
   ** Post:     The number of bits that code for FSM internal states
   **           is set to b, so that the number of these states is 2^b.
   */
//  static int    FSMSize (void);
   /*
   **  Pre:     True.
   **
   ** Post:     Returns FSM size as 2 raised to the power of
   **           [m+b], where m=FSM memory and b=number of bits
   **           that code for the FSM internal states.
   */
 // virtual void  InitProgram (void);   // Called by constructor
   /*
   **  Pre:     The number of bits that code for the FSM internal
   **           states cannot exceed 15.
   **
   ** Post: (1) This tradebot constructs a new FSM look-up table with
   **           memory and internal states determined in accordance
   **           with tng.ini specifications and with 1 output bit
   **           interpreted as 0=defect and 1=cooperate.
   **       (2) This tradebot initializes its FSM look-up table with
   **           an array of integers extracted from its genotype.
   */
//  virtual void  Init (tngSimulation* sim, bioList& traders,
//             int maxOffers, int maxAccepts);
   /*
   **  Pre:     The parameter values set in tng.ini are admissible.
   **
   ** Post: (1) This tradebot initializes its member data with values
   **           from tng.ini.
   **       (2) This tradebot constructs an array for holding
   **           information about all other tradebots.
   **       (3) If this tradebot is a seller, it constructs another
   **           array to use as a waiting list for holding offers
   **           received from buyer tradebots.
   **       (4) This tradebot assigns an expected payoff -1 to itself
   **           (so it will never trade with itself) and an expected
   **           payoff to each other tradebot in accordance with the
   **           specifications in tng.ini.
   **       (5) Each tradebot is assigned a name in accordance with
   **           its type (PureBuyer, PureSeller, or BuyerSeller).
   **           and a number to identify it within its type.
   */
//  virtual void  Look (void) {}
   /*
   **  Null definition for the pure virtual function Look() in bioPtype.
   */
//  virtual void  PrepareOffers (void);
  //: Preparation of a tradebot for entering into a trade cycle.
   /*
   **  Pre:      The parameter values set in tng.ini are admissible.
   **
   ** Post:  (1) The offer status of each other tradebot in relation
   **            to this tradebot is set to NO_OFFER.
   **        (2) The number of trade offers this tradebot currently
   **            has outstanding with other tradebots is set to 0.
   **        (3) The number of trade offers this tradebot has made is
   **            set to 0.
   */
// virtual int    SubmitOffers (void);
  //: How a buyer submits new trade offers in any given trade cycle.
   /*
   **  Pre:  (1) This tradebot is a buyer.
   **        (2) The parameter values set in tng.ini are admissible.
   **
   ** Post:  (1) 0 is returned if the number of this tradebot's
   **            currently outstanding trade offers equals this
   **            tradebot's buyer offer quota.
   **        (2) If this tradebot's outstanding trade offers are less
   **            than its buyer offer quota:
   **            (a) This tradebot sorts all seller tradebots
   **                from highest to lowest expected payoff.
   **            (b) This tradebot makes new trade offers to the
   **                tolerable sellers it most prefers to whom it
   **                has not previously made an offer, subject to the
   **                constraint that its outstanding trade offers
   **                cannot exceed its buyer offer quota.
   **            (c) This tradebot changes the status of each seller
   **                to whom it makes a new trade offer to OFFER_MADE.
   **            (d) The number of new trade offers is returned.
   */
//  virtual void  TakeOffer (tngTradeBot* bot);
  //: Receipt of a buyer's offer by a seller.
   /*
   **  Pre:  (1) This tradebot is a seller.
   **        (2) bot is a buyer.
   **
   ** Post:      This tradebot provisionally adds a trade offer
   **            from bot to its waiting list.
   */
 // virtual void  InactiveCheck (void) ;
  //:  Is this tradebot inactive?
   /*
   **  Pre:  (1) The number of offers this tradebot made during
   **            the last trade cycle is recorded.
   **        (2) The number of offers this tradebot accepted during
   **            the last trade cycle is recorded.
   **
   ** Post:  (1) This tradebot is tested to see if it was inactive
   **            in the last trade cycle, meaning it neither made nor
   **            accepted any trade offers.
   **        (2) This tradebot, if inactive, adds inactive
   **            payoff to its total payoff sum.
   **        (3) This tradebot, if inactive, adds 1 to its total
   **            payoff count and 1 to its total inactive count.
   */
//  virtual void  AcceptOffers (void);
  //: Provisional acceptance of best buyer offers by a seller.
   /*
   **  Pre:      This tradebot is a seller.
   **
   ** Post:  (1) This tradebot sorts the trade offers currently on
   **            its waiting list on the basis of its expected payoff
   **            assessments for the buyers who submitted the offers.
   **        (2) This tradebot provisionally accepts the most preferred
   **            tolerable trade offers currently on its waiting list,
   **            subject to the constraint that the number of its accepted
   **            trade offers cannot exceed its seller acceptance quota.
   **        (3) This tradebot refuses all other trade offers currently
   **            on its waiting list.
   */
//  virtual void  OfferRejected (tngTradeBot* bot);
  //:  A buyer's response when a seller refuses its trade offer.
   /*
   **  Pre:  (1) This tradebot is a buyer.
   **        (2) bot is a seller.
   **        (3) bot has just refused a trade offer from this tradebot.
   **
   ** Post:  (1) This tradebot adds a refusal payoff to its total
   **            payoff sum.
   **        (2) This tradebot adds 1 to its total payoff count.
   **        (3) This tradebot adds a refusal payoff to its payoff
   **            sum from previous interactions with bot.
   **        (4) This tradebot adds 1 to its payoff count with bot
   **            and 1 to its refusal count with bot.
   **        (5) This tradebot updates its expected payoff for bot.
   **        (6) This tradebot changes bot's status to OFFER_REJECTED.
   **        (7) This tradebot decrements its outstanding offers by 1.
   */
//  virtual void  PlayPD (void);
  //: Engage in trade. This message is sent to sellers at the end of the
  //: matching process, after new trade offers from buyers have ceased.
   /*
   **  Pre:      This tradebot is a seller.
   **
   ** Post:      This tradebot engages in one trade (PD game) with
   **            each buyer who has a trade offer on its current
   **            waiting list that has not been refused.
   */
//  virtual int  GetAction (tngTradeBot* bot);
  //: What action should this tradebot take in its next trade with bot?
   /*
   **  Pre:  (1) This tradebot has an initialized FSM.
   **        (2) This tradebot has recorded the FSM state it is in
   **            with respect to bot (if any) and the last action
   **            (if any) that bot took in a trade with this tradebot.
   **
   ** Post:     This tradebot determines and returns the action it
   **           will take in a next trade with bot, either 0=defect
   **           or 1=cooperate, and updates its current FSM state
   **           with respect to bot.
   */
//  virtual void  Buy (tngTradeBot* bot, float lastPayoff, int action);
   //: How a buyer updates its information on a seller after a trade.
   /*
   **  Pre:  (1) This tradebot is a buyer.
   **        (2) bot is a seller.
   **        (3) This tradebot received lastPayoff in a trade with bot.
   **        (4) action was bot's action in this trade.
   **
   ** Post:  (1) This tradebot adds lastPayoff to its total payoff sum.
   **        (2) This tradebot adds 1 to its total payoff count.
   **        (3) This tradebot adds lastPayoff to its payoff sum from
   **            previous interactions with bot.
   **        (4) This tradebot adds 1 to its payoff count with bot.
   **        (5) This tradebot records action.
   **        (6) This tradebot adds 1 to its trade count with bot.
   **        (7) This tradebot updates its expected payoff for bot.
   */
 // virtual void  Sell (tngTradeBot* bot, float lastPayoff, int action);
   //: How a seller updates its information on a buyer after a trade.
   /*
   **  Pre:  (1) This tradebot is a seller.
   **        (2) bot is a buyer.
   **        (3) This tradebot received lastPayoff in a trade with bot.
   **        (4) action was bot's action in this trade.
   **
   ** Post:  (1) This tradebot adds lastPayoff to its total payoff sum.
   **        (2) This tradebot adds 1 to its total payoff count.
   **        (3) This tradebot adds lastPayoff to its payoff sum from
   **            previous interactions with bot.
   **        (4) This tradebot adds 1 to its payoff count with bot.
   **        (5) This tradebot records action.
   **        (6) This tradebot adds 1 to its trade count with bot.
   **        (7) This tradebot updates its expected payoff for bot.
   */
//  virtual void  UpdateExp (TraderInfo* info);
  //: How should this tradebot update its expected payoff assessment
  //: for a tradebot it has just interacted with?
   /*
   **  Pre:  (1) This tradebot has just had an interaction with
   **            the bot whose information is recorded in *info.
   **        (2) This tradebot has recorded the payoff it received
   **            from this last interaction with bot.
   **        (3) This tradebot has incremented its payoff count with
   **            bot by 1 to reflect this last interaction.
   **
   ** Post:      The expected payoff assessment of this tradebot for
   **            bot is updated to reflect the outcome of their last
   **            interaction.
   */
//  virtual float  CalcFitness (void);
   //: Calculate this tradebot's fitness score.
   /*
   **  Pre:  (1) This tradebot has recorded its total payoff sum and
   **            its total payoff count.
   **        (2) This tradebot's total payoff count is positive.
   **
   **  Post:     Returns this tradebot's fitness score, calculated as
   **            its total payoff sum divided by its total payoff count.
   */
//  virtual void  Dump (void);
  //: Print info about this tradebot and about the info this
  //: tradebot has recorded about other tradebots.
   /*
   **  Pre:      True
   **
   ** Post:  (1) This tradebot prints its identifier, its total
   **            payoff sum, its total number of inactive payoffs,
   **            and its fitness score.
   **        (2) This tradebot prints info that it has recorded for each
   **            other bot, as follows: the bot's list number; the bot's
   **            identifier; this tradebot's payoff sum from interactions
   **            with bot; this tradebot's refusal count with bot; the
   **            expected payoff that this tradebot currently associates
   **            with bot; this tradebot's trade count with bot; and the
   **            actions bot has taken in previous trades with this
   **            tradebot.
   */
//  virtual TraderInfo*  GetTraderInfo (tngTradeBot* bot);
   /*
   **  Pre:      True
   **
   ** Post:      This tradebot retrieves and returns a pointer to
   **            info that it has recorded for bot.
   */
#endif  // TNGBOT_H





